from dataclasses import dataclass
from typing import Callable
from reactivex import ConnectableObservable, Observable
from reactivex.disposable import CompositeDisposable
from reactivex.scheduler import ThreadPoolScheduler

from ccxt import luno

from elm_framework_helpers.websockets.models import WebsocketBundle

from bittrade_luno.models.enhanced_websocket import EnhancedWebsocket
from bittrade_luno.models.rest.cancel_order import CancelOrderRequest, CancelOrderResponse
from bittrade_luno.models.rest.create_order import LimitOrderRequest, LimitOrderResponse
from bittrade_luno.models.rest.list_trade import ListTradeRequest, ListTradeResponse
from bittrade_luno.models.rest.order_book import ListOrderBookResponse, OrderBookRequest


@dataclass
class FrameworkContext:
    all_subscriptions: CompositeDisposable
    exchange: luno
    scheduler: ThreadPoolScheduler
    stream_bundles: ConnectableObservable[WebsocketBundle]
    stream_sockets: Observable[EnhancedWebsocket]
    stream_socket_messages: Observable[dict]
    private_stream_bundles: ConnectableObservable[WebsocketBundle]
    private_stream_sockets: Observable[EnhancedWebsocket]
    private_stream_socket_messages: Observable[dict]
    list_trade_http: Callable[[ListTradeRequest], Observable[ListTradeResponse]]
    list_top_orderbook_http: Callable[[OrderBookRequest], Observable[ListOrderBookResponse]]
    create_order_http: Callable[[LimitOrderRequest], Observable[LimitOrderResponse]]
    cancel_order_http: Callable[[CancelOrderRequest], Observable[CancelOrderResponse]]



from enum import Enum


class LunoEndpoints(Enum):
    POST_LIMIT_ORDER = "/api/1/postorder"
    CANCEL_ORDER = "/api/1/stoporder"
    LIST_RECENT_TRADES = "/api/1/trades"
    TOP_ORDER_BOOK = "/api/1/orderbook_top"



from bittrade_luno.models import endpoints
from bittrade_luno.models.request_message import RequestMessage
from bittrade_luno.models.rest.order_book import ListOrderBookResponse, OrderBookRequest
from bittrade_luno.rest.http_factory_decorator import http_factory


@http_factory(ListOrderBookResponse)
def list_top_orderbook_http_factory(
    params: OrderBookRequest,
):
    return RequestMessage(
        method="GET",
        endpoint=endpoints.LunoEndpoints.TOP_ORDER_BOOK,
        params=params.to_dict(),
    )


from bittrade_luno.models import endpoints
from bittrade_luno.models.request_message import RequestMessage
from bittrade_luno.models.rest.create_order import LimitOrderRequest, LimitOrderResponse
from bittrade_luno.rest.http_factory_decorator import http_factory


@http_factory(LimitOrderResponse)
def create_order_http_factory(
    params: LimitOrderRequest,
):
    return RequestMessage(
        method="POST",
        endpoint=endpoints.LunoEndpoints.POST_LIMIT_ORDER,
        params=params.to_dict(),
    )


import requests
from logging import getLogger
from os import getenv
from reactivex import Observable, abc
from reactivex.disposable import Disposable

from bittrade_luno.models.request_message import RequestMessage

logger = getLogger(__name__)

LUNO_HTTP_URL = getenv("LUNO_HTTP_URL", "https://api.luno.com")
session = requests.Session()

def prepare_request(message: RequestMessage) -> requests.models.Request:
    http_method = message.method
    kwargs = {}
    # check if message params are set, if not, ignores
    if message.params:
        if http_method == "GET":
            kwargs["params"] = message.params
        else:
            kwargs["data"] = message.params

    # There are (few) cases where the endpoint must be a string; "handle" that below
    try:
        endpoint = message.endpoint.value
    except AttributeError:
        endpoint = message.endpoint
    return requests.Request(http_method, f"{LUNO_HTTP_URL}{endpoint}", **kwargs)


def send_request(request: requests.models.Request) -> Observable:
    def subscribe(
        observer: abc.ObserverBase,
        scheduler: abc.SchedulerBase | None = None,
    ) -> abc.DisposableBase:
        response = session.send(request.prepare())
        if response.ok:
            try:
                body = response.json()
                observer.on_next(body)
                observer.on_completed()
            except Exception as exc:
                logger.error(
                    "Error parsing request %s; request was %s",
                    response.text,
                    response.request.body
                    if request.method == "POST"
                    else response.request.headers,
                )
                observer.on_error(exc)
        else:
            try:
                logger.error(
                    "Error with request %s; request was %s",
                    response.text,
                    response.request.body
                    if request.method == "POST"
                    else response.request.headers,
                )
                response.raise_for_status()
            except Exception as exc:
                observer.on_error(exc)
        return Disposable()

    return Observable(subscribe)


from logging import getLogger
from typing import Any, Optional

import orjson
import reactivex.disposable
from reactivex import Observable
from reactivex.abc import SchedulerBase, ObserverBase
from reactivex.scheduler import ThreadPoolScheduler
from websocket import WebSocketConnectionClosedException, WebSocketApp
from elm_framework_helpers.websockets.models import WebsocketBundle, WEBSOCKET_MESSAGE, WEBSOCKET_HEARTBEAT, WEBSOCKET_STATUS, WEBSOCKET_OPENED, WEBSOCKET_CLOSED
from bittrade_luno.models.enhanced_websocket import EnhancedWebsocket

logger = getLogger(__name__)
raw_logger = getLogger("bittrade_luno.raw_socket.received")


def raw_websocket_connection(
    url: str, scheduler: Optional[SchedulerBase] = None
) -> Observable[WebsocketBundle]:
    def subscribe(
        observer: ObserverBase[WebsocketBundle],
        scheduler_: Optional[SchedulerBase] = None,
    ):
        _scheduler = scheduler or scheduler_ or ThreadPoolScheduler()
        connection: WebSocketApp | None = None

        def action(*args: Any):
            nonlocal connection

            def on_error(_ws: WebSocketApp, error: Exception):
                logger.error("[SOCKET][RAW] Websocket errored %s", error)
                observer.on_next(
                    (enhanced, WEBSOCKET_STATUS, WEBSOCKET_CLOSED))
                observer.on_error(error)

            def on_close(_ws: WebSocketApp, close_status_code: int, close_msg: str):
                logger.warning(
                    "[SOCKET][RAW] Websocket closed | url: %s, status: %s, close message: %s",
                    url,
                    close_status_code,
                    close_msg,
                )
                observer.on_next(
                    (enhanced, WEBSOCKET_STATUS, WEBSOCKET_CLOSED))
                observer.on_error(Exception("Socket closed"))

            def on_open(_ws: WebSocketApp):
                logger.info("[SOCKET][RAW] Websocket opened at %s", url)
                observer.on_next(
                    (enhanced, WEBSOCKET_STATUS, WEBSOCKET_OPENED))

            def on_message(ws: WebSocketApp, message: bytes | str):
                pass_message = orjson.loads(message)
                raw_logger.debug(message)
                logger.debug("[SOCKET][RAW] %s", message)

                try:
                    observer.on_next((enhanced, WEBSOCKET_MESSAGE, pass_message))
                except:
                    logger.exception("[SOCKET] Error on socket message")

            connection = WebSocketApp(
                url,
                on_open=on_open,
                on_close=on_close,
                on_error=on_error,
                on_message=on_message,
            )
            enhanced = EnhancedWebsocket(connection)

            def run_forever(*args: Any):
                assert connection is not None
                connection.run_forever()

            _scheduler.schedule(run_forever)

        def disconnect():
            logger.info("[SOCKET] Releasing resources")
            assert connection is not None
            try:
                connection.close()
            except WebSocketConnectionClosedException as exc:
                logger.error("[SOCKET] Socket was already closed %s", exc)

        return reactivex.disposable.CompositeDisposable(
            _scheduler.schedule(
                action), reactivex.disposable.Disposable(disconnect)
        )

    return Observable(subscribe)


__all__ = ["raw_websocket_connection"]



from typing import Callable, Optional

from reactivex import ConnectableObservable
from reactivex.operators import publish, map
from reactivex.abc import SchedulerBase
import os
from elm_framework_helpers.websockets.models import WebsocketBundle

from bittrade_luno.connection.generic import raw_websocket_connection
from bittrade_luno.connection.reconnect import retry_with_backoff


LUNO_STREAM_URL = os.getenv('LUNO_PUBLIC_WEBSOCKET', 'wss://ws.luno.com/api/1/stream/')

def public_websocket_connection(
    *, pair: str, reconnect: bool = True, scheduler: Optional[SchedulerBase] = None
) -> ConnectableObservable[WebsocketBundle]:
    url = f"{LUNO_STREAM_URL}{pair}"
    connection = raw_websocket_connection(url, scheduler=scheduler)
    if reconnect:
        connection = connection.pipe(retry_with_backoff)
    return connection.pipe(publish())

__all__ = [
    "public_websocket_connection",
]



import reactivex

from elm_framework_helpers.operators.retry_with_delay import resettable_counter, retry_with_delay

retry_with_backoff = retry_with_delay(resettable_counter([0, 0, 1, 2, 5], infinite_behavior="loop"), reset_after=reactivex.empty())

__all__ = [
    "retry_with_backoff",
]


from dataclasses import asdict, dataclass
from typing import TypedDict


@dataclass
class LimitOrderRequest:
    pair: str
    type: str
    post_only: bool
    volume: str
    price: str
    time_in_force: str = "GTC"
    stop_price: str = None
    stop_direction: str = None
    base_account_id: int = None
    counter_account_id: int = None
    timestamp: int = None
    ttl: int = None
    client_order_id: str = None


    def to_dict(self):
        as_dict = asdict(self)
        return as_dict

class LimitOrderResponse(TypedDict):
    order_id: str


import functools
from typing import Any, Callable, ParamSpec, Type, TypeVar, TypedDict, cast
import requests
from reactivex import Observable
from bittrade_luno.connection import http

from bittrade_luno.models.request_message import RequestMessage

P = ParamSpec("P")


# TODO this typing does not work, it does not allow us to define the sub type of the response's result
R = TypeVar("R")


def http_factory(return_type: Type):
    def factory_wrapper(
        fn: Callable[P, RequestMessage]
    ) -> Callable[
        [Callable[[requests.models.Request], requests.models.Request]],
        Callable[P, Observable[return_type]],
    ]:
        @functools.wraps(fn)
        def factory(
            add_token: Callable[[requests.models.Request], requests.models.Request]
        ):
            def inner(*args: P.args, **kwargs: P.kwargs) -> Observable[return_type]:
                request = fn(*args, **kwargs)
                return cast(
                    Observable[return_type],
                    http.send_request(add_token(http.prepare_request(request))),
                )

            return inner

        return factory

    return factory_wrapper



import dataclasses
import time
from typing import Any, Literal

from bittrade_luno.models import endpoints


@dataclasses.dataclass(frozen=True)
class RequestMessage:
    method: Literal["GET", "POST", "PUT", "DELETE"]
    endpoint: endpoints.LunoEndpoints
    params: dict[str, Any] = dataclasses.field(default_factory=dict)
    client_order_id: int = 0



from dataclasses import asdict, dataclass
from decimal import Decimal
from typing import List, TypedDict


@dataclass
class ListTradeRequest:
    pair: str
    since: int

    def to_dict(self):
        as_dict = asdict(self)
        return as_dict

@dataclass
class Trade:
    is_buy: bool
    price: str
    sequence: int
    timestamp: int
    volume: str

class TradeResponse(TypedDict):
    is_buy: bool
    price: str
    sequence: int
    timestamp: int
    volume: str
    
class ListTradeResponse(TypedDict):
    trades: List[TradeResponse]


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bittrade_luno',
 'bittrade_luno.connection',
 'bittrade_luno.framework',
 'bittrade_luno.models',
 'bittrade_luno.models.rest',
 'bittrade_luno.rest']

package_data = \
{'': ['*']}

install_requires = \
['asyncio>=3.4.3,<4.0.0',
 'ccxt==2.6.5',
 'elm-framework-helpers>=0.2.6,<0.3.0',
 'expression>=4.2.4,<5.0.0',
 'orjson>=3.9.4,<4.0.0',
 'prompt-toolkit>=3.0.39,<4.0.0',
 'ptpython>=3.0.23,<4.0.0',
 'pydantic==1.10.12',
 'reactivex>=4.0.4,<5.0.0',
 'requests>=2.31.0,<3.0.0',
 'rich>=13.5.2,<14.0.0',
 'websocket-client>=1.6.1,<2.0.0',
 'websockets>=10.4,<11.0']

setup_kwargs = {
    'name': 'bittrade-luno',
    'version': '0.1.0',
    'description': 'Reactive for luno',
    'long_description': '# bittrade luno\n> rxpy based LUNO stream and rest api',
    'author': 'Yu Jin',
    'author_email': 'adamlyj89@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



from typing import Any
from elm_framework_helpers.websockets import models
import orjson


class EnhancedWebsocket(models.EnhancedWebsocket):
    def send_message(self, message: Any) -> int | str:
        return self.send_json(message)

    def prepare_request(self, message: Any) -> tuple[str, bytes]:
        self._id += 1
        return f"id{self._id}", orjson.dumps(message)



from dataclasses import asdict, dataclass
from typing import TypedDict


@dataclass
class CancelOrderRequest:
    order_id: str

    def to_dict(self):
        as_dict = asdict(self)
        return as_dict

class CancelOrderResponse(TypedDict):
    success: bool


from bittrade_luno.models import endpoints
from bittrade_luno.models.request_message import RequestMessage
from bittrade_luno.models.rest.cancel_order import CancelOrderRequest, CancelOrderResponse
from bittrade_luno.rest.http_factory_decorator import http_factory


@http_factory(CancelOrderResponse)
def cancel_order_http_factory(
    params: CancelOrderRequest,
):
    return RequestMessage(
        method="POST",
        endpoint=endpoints.LunoEndpoints.CANCEL_ORDER,
        params=params.to_dict(),
    )


from dataclasses import asdict, dataclass
from decimal import Decimal
from typing import List, TypedDict


@dataclass
class OrderBookRequest:
    pair: str

    def to_dict(self):
        as_dict = asdict(self)
        return as_dict

class OrderBookResponse(TypedDict):
    price: str
    volume: str
    
class ListOrderBookResponse(TypedDict):
    asks: List[OrderBookResponse]
    bids: List[OrderBookResponse]
    timestamp: int


from bittrade_luno.models import endpoints
from bittrade_luno.models.request_message import RequestMessage
from bittrade_luno.models.rest.list_trade import ListTradeRequest, ListTradeResponse
from bittrade_luno.rest.http_factory_decorator import http_factory


@http_factory(ListTradeResponse)
def list_trade_http_factory(
    params: ListTradeRequest,
):
    return RequestMessage(
        method="GET",
        endpoint=endpoints.LunoEndpoints.LIST_RECENT_TRADES,
        params=params.to_dict(),
    )


import requests
from typing import Callable, cast
from ccxt import luno
from reactivex import Observable, operators
from reactivex.disposable import CompositeDisposable
from reactivex.scheduler import ThreadPoolScheduler
from reactivex.subject import BehaviorSubject
from elm_framework_helpers.websockets.operators import connection_operators
from elm_framework_helpers.websockets.models.bundle import WebsocketBundle

from bittrade_luno.connection.private import private_websocket_connection
from bittrade_luno.connection.public import public_websocket_connection
from bittrade_luno.models.enhanced_websocket import EnhancedWebsocket
from bittrade_luno.models.framework import FrameworkContext
from bittrade_luno.rest.cancel_order import cancel_order_http_factory
from bittrade_luno.rest.create_order import create_order_http_factory
from bittrade_luno.rest.list_trades import list_trade_http_factory
from bittrade_luno.rest.top_orderbook import list_top_orderbook_http_factory

def get_framework(
    *,
    pair: str,
    authenticate_signer: Callable[
        [EnhancedWebsocket], EnhancedWebsocket
    ] = None, 
    trade_signer_http: Callable[
        [requests.models.Request], requests.models.Request
    ] = None,
    load_markets=True,
) -> FrameworkContext:
    exchange = luno()
    if load_markets:
        exchange.load_markets()

    details = exchange.market(pair)
    pair = details.get("id")

    pool_scheduler = ThreadPoolScheduler(200)
    all_subscriptions = CompositeDisposable()

    stream_bundles = public_websocket_connection(
        pair=pair
    )
    stream_sockets = stream_bundles.pipe(
        connection_operators.keep_new_socket_only(),
        operators.share(),
    )
    stream_socket_messages = stream_bundles.pipe(
        connection_operators.keep_messages_only(), operators.share()
    )
    stream_sockets.subscribe(authenticate_signer)

    private_stream_bundles = private_websocket_connection()
    private_stream_sockets = private_stream_bundles.pipe(
        connection_operators.keep_new_socket_only(),
        operators.share(),
    )
    private_stream_socket_messages = private_stream_bundles.pipe(
        connection_operators.keep_messages_only(), operators.share()
    )
    private_stream_sockets.subscribe(authenticate_signer)

    list_trade_http = list_trade_http_factory(trade_signer_http)
    list_top_orderbook_http = list_top_orderbook_http_factory(trade_signer_http)
    create_order_http = create_order_http_factory(trade_signer_http)
    cancel_order_http = cancel_order_http_factory(trade_signer_http)

    return FrameworkContext(
        all_subscriptions=all_subscriptions,
        exchange=exchange,
        scheduler=pool_scheduler,
        stream_bundles=stream_bundles,
        stream_sockets=stream_sockets,
        stream_socket_messages=stream_socket_messages,
        private_stream_bundles=private_stream_bundles,
        private_stream_sockets=private_stream_sockets,
        private_stream_socket_messages=private_stream_socket_messages,
        list_trade_http=list_trade_http,
        list_top_orderbook_http=list_top_orderbook_http,
        create_order_http=create_order_http,
        cancel_order_http=cancel_order_http,
    )


